from .fake import FakeDataSource
from .github import GithubDataSource
from .googlesheets import GoogleSheetsDataSource
from .huggingface import HuggingFaceDatasets
from .kaggle import KaggleDataSource
from .simplejson import SimpleJsonDataSource
from .stock import StockDataSource
